# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals, division
from eight import *

from .lca import LCA
from .utils import wrap_functional_unit
try:
    from bw2data import calculation_setups
except ImportError:
    calculation_setups = None
import numpy as np


class MultiLCA(object):
    """Wrapper class for performing LCA calculations with many functional units and LCIA methods.

    Needs to be passed a ``calculation_setup`` name.

    This class does not subclass the `LCA` class, and performs all calculations upon instantiation.

    Initialization creates `self.results`, which is a NumPy array of LCA scores, with rows of functional units and columns of LCIA methods. Ordering is the same as in the `calculation_setup`.

    """
    def __init__(self, cs_name, log_config=None):
        if calculation_setups is None:
            raise ImportError
        assert cs_name in calculation_setups
        try:
            cs = calculation_setups[cs_name]
        except KeyError:
            raise ValueError(
                "{} is not a known `calculation_setup`.".format(cs_name)
            )
        self.func_units = cs['inv']
        self.methods = cs['ia']
        self.lca = LCA(demand=self.all, method=self.methods[0], log_config=log_config)
        self.lca.logger.info({
            'message': 'Started MultiLCA calculation',
            'methods': list(self.methods),
            'functional units': [wrap_functional_unit(o) for o in self.func_units]
        })
        self.lca.lci(factorize=True)
        self.method_matrices = []
        self.results = np.zeros((len(self.func_units), len(self.methods)))
        for method in self.methods:
            self.lca.switch_method(method)
            self.method_matrices.append(self.lca.characterization_matrix)

        for row, func_unit in enumerate(self.func_units):
            self.lca.redo_lci(func_unit)
            for col, cf_matrix in enumerate(self.method_matrices):
                self.lca.characterization_matrix = cf_matrix
                self.lca.lcia_calculation()
                self.results[row, col] = self.lca.score

    @property
    def all(self):
        """Get all possible databases by merging all functional units"""
        return {key: 1 for func_unit in self.func_units for key in func_unit}



from io import BytesIO

# import numpy as np

from bw2calc.utils import get_seed, wrap_functional_unit, os, load_arrays, np
import multiprocessing
import pytest
import sys

MAX_INT_32 = 4294967295


@pytest.mark.skipif(sys.version_info < (3,0), reason="MP pool changes")
def test_get_seeds_different_under_mp_pool():
    with multiprocessing.Pool(processes=4) as pool:
        results = list(pool.map(get_seed, [None] * 10))
    assert sorted(set(results)) == sorted(results)


def test_wrap_functional_unit():
    given = {17: 42}
    expected = {'key': 17, 'amount': 42}
    assert wrap_functional_unit(given) == [expected]

    given = {('a', 'b'): 42}
    expected = {'database': 'a', 'code': 'b', 'amount': 42}
    assert wrap_functional_unit(given) == [expected]

    class Foo:
        def __getitem__(self, index):
            if index == 0:
                return 'a'
            elif index == 1:
                return 'b'

    given = {Foo(): 42}
    expected = {'database': 'a', 'code': 'b', 'amount': 42}
    assert wrap_functional_unit(given) == [expected]


def test_load_arrays_order(monkeypatch):
    monkeypatch.setattr(np, "load", lambda x, allow_pickle: x)
    monkeypatch.setattr(np, "hstack", lambda x: x)
    given = [1, "a", np.ones(5), 2, "b", np.zeros(3)]
    result = load_arrays(given)
    assert result[0].sum() == 5
    assert result[1].sum() == 0
    assert result[2:] == ["a", "b", 1, 2]


def test_load_arrays_string():
    # use the independent lca fixture data
    fpath = os.path.join(os.path.dirname(__file__), "fixtures", "independent", "ia.npy")

    dtype = [
        ('flow', np.uint32),
        ('row', np.uint32),
        ('col', np.uint32),
        ('geo', np.uint32),
        ('amount', np.float32),
    ]
    expected_array = np.asarray(
        [(10, MAX_INT_32, MAX_INT_32, 17, 1),
         (11, MAX_INT_32, MAX_INT_32, 17, 10)],
        dtype=dtype)

    actual_array = load_arrays([fpath])

    assert np.all(actual_array == expected_array), "Failed loading array from filepath"


def test_load_arrays_bytesio():
    dtype = [
        ('flow', np.uint32),
        ('row', np.uint32),
        ('col', np.uint32),
        ('geo', np.uint32),
        ('amount', np.float32),
    ]
    expected_array = np.asarray(
        [(10, MAX_INT_32, MAX_INT_32, 17, 1),
         (11, MAX_INT_32, MAX_INT_32, 17, 10.2)],
        dtype=dtype)

    with BytesIO() as b:
        np.save(b, expected_array)
        b.seek(0)  # return to beginning of file
        actual_array = load_arrays([b])

    assert np.all(actual_array == expected_array), "Failed loading array from bytes"




# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
from eight import *

from .fixtures.presamples_basic import write_database, basedir
from bw2calc.errors import OutsideTechnosphere, NonsquareTechnosphere
from bw2calc import (
    ComparativeMonteCarlo,
    LCA,
    MonteCarloLCA,
    ParameterVectorLCA,
)
from bw2calc.lca import PackagesDataLoader
from bw2data import *
from bw2data.tests import bw2test
from bw2data.utils import TYPE_DICTIONARY
import numpy as np
import pytest
import os

try:
    from pathlib import Path
except ImportError:
    Path = None


@pytest.fixture
@bw2test
def basic():
    write_database()


def test_writing_test_fixture(basic):
    assert len(databases) == 2
    assert len(methods) == 1
    lca  = LCA({("test", "1"): 1})
    lca.lci()
    expected = [
        (("test", "1"), ("test", "1"), 1),
        (("test", "3"), ("test", "1"), -0.1),
        (("test", "2"), ("test", "2"), 0.5),
        (("test", "1"), ("test", "2"), 2),
        (("test", "3"), ("test", "3"), 1),
        (("test", "1"), ("test", "3"), -3),
        (("test", "2"), ("test", "3"), -2),

    ]
    for x, y, z in expected:
        assert np.allclose(lca.technosphere_matrix[
            lca.product_dict[x], lca.activity_dict[y]
        ], z)
    expected = [
        (("bio", "b"), ("test", "1"), 7),
        (("bio", "a"), ("test", "2"), 1),
        (("bio", "b"), ("test", "2"), 5),
        (("bio", "a"), ("test", "3"), 2),
    ]
    for x, y, z in expected:
        assert np.allclose(lca.biosphere_matrix[
            lca.biosphere_dict[x] , lca.activity_dict[y]
        ], z)

@pytest.mark.skipif(not PackagesDataLoader, reason="presamples not installed")
@pytest.mark.skipif(not Path, reason="Python version 2")
def test_accept_pathlib(basic):
    ss = Path(basedir)  / "single-sample"
    lca = LCA({("test", "2"): 1}, method=("m",), presamples=[ss])
    lca.lci()

@pytest.mark.skipif(not PackagesDataLoader, reason="presamples not installed")
def test_single_sample_presamples(basic):
    ss = os.path.join(basedir, "single-sample")

    lca = LCA({("test", "2"): 1}, method=("m",))
    lca.lci()
    assert np.allclose(
        lca.supply_array,
        np.array([-(2 + 2/3),      14/15, -4/15])
    )
    lca = LCA({("test", "2"): 1}, method=("m",), presamples=[ss])
    lca.lci()
    assert np.allclose(
        lca.supply_array,
        np.array([2, 1.4, 0.2])
    )

    mc = MonteCarloLCA({("test", "2"): 1}, method=("m",))
    next(mc)
    assert np.allclose(
        mc.supply_array,
        np.array([-(2 + 2/3),      14/15, -4/15])
    )
    mc = MonteCarloLCA({("test", "2"): 1}, method=("m",), presamples=[ss])
    next(mc)
    assert np.allclose(
        mc.supply_array,
        np.array([2, 1.4, 0.2])
    )

    mc = ParameterVectorLCA({("test", "2"): 1}, method=("m",))
    next(mc)
    assert mc.technosphere_matrix[
        mc.product_dict[("test", "2")],
        mc.activity_dict[("test", "2")],
    ] == 0.5
    mc = ParameterVectorLCA({("test", "2"): 1}, method=("m",), presamples=[ss])
    next(mc)
    assert mc.technosphere_matrix[
        mc.product_dict[("test", "2")],
        mc.activity_dict[("test", "2")],
    ] == 1

    mc = ComparativeMonteCarlo([{("test", "2"): 1}], method=("m",))
    next(mc)
    assert mc.technosphere_matrix[
        mc.product_dict[("test", "2")],
        mc.activity_dict[("test", "2")],
    ] == 0.5
    mc = ComparativeMonteCarlo([{("test", "2"): 1}], method=("m",),
                               presamples=[ss])
    next(mc)
    assert mc.technosphere_matrix[
        mc.product_dict[("test", "2")],
        mc.activity_dict[("test", "2")],
    ] == 1

@pytest.mark.skipif(not PackagesDataLoader, reason="presamples not installed")
def test_solver_cache_invalidated(basic):
    ss = os.path.join(basedir, "single-sample")
    lca = LCA({("test", "2"): 1}, method=("m",), presamples=[ss])
    lca.lci(factorize=True)
    assert hasattr(lca, "solver")
    lca.presamples.update_matrices()
    assert not hasattr(lca, "solver")

@pytest.mark.skipif(not PackagesDataLoader, reason="presamples not installed")
def test_multi_sample_presamples(basic):
    path = os.path.join(basedir, "multi")

    lca = LCA({("test", "2"): 1}, method=("m",))
    lca.lci()
    static = lca.technosphere_matrix.data

    multi = []
    for _ in range(10):
        lca = LCA({("test", "2"): 1}, method=("m",), presamples=[path])
        lca.lci()
        multi.append(lca.technosphere_matrix.data)

    assert all(np.allclose(multi[i], multi[i + 1]) for i in range(9))
    for x in range(9):
        assert not np.allclose(multi[x], static)

@pytest.mark.skipif(not PackagesDataLoader, reason="presamples not installed")
def test_multi_sample_presamples_no_seed_different(basic):
    path = os.path.join(basedir, "unseeded")

    multi = []
    for _ in range(10):
        lca = LCA({("test", "2"): 1}, method=("m",), presamples=[path])
        lca.lci()
        multi.append(lca.technosphere_matrix.data)

    assert not all(np.allclose(multi[i], multi[i + 1]) for i in range(9))

@pytest.mark.skipif(not PackagesDataLoader, reason="presamples not installed")
def test_keep_presamples_seed(basic):
    path = os.path.join(basedir, "multi")

    mc = MonteCarloLCA({("test", "2"): 1}, method=("m",), presamples=[path], seed=6)
    first = [next(mc) for _ in range(10)]
    mc = MonteCarloLCA({("test", "2"): 1}, method=("m",), presamples=[path], seed=7)
    second = [next(mc) for _ in range(10)]
    assert first == second

@pytest.mark.skipif(not PackagesDataLoader, reason="presamples not installed")
def test_override_presamples_seed(basic):
    path = os.path.join(basedir, "multi")

    mc = MonteCarloLCA({("test", "2"): 1}, method=("m",), presamples=[path], seed=6, override_presamples_seed=True)
    first = [next(mc) for _ in range(10)]
    mc = MonteCarloLCA({("test", "2"): 1}, method=("m",), presamples=[path], seed=7, override_presamples_seed=True)
    second = [next(mc) for _ in range(10)]
    assert first != second

@pytest.mark.skipif(not PackagesDataLoader, reason="presamples not installed")
def test_sequential_seed():
    path = os.path.join(basedir, "seq")
    lca = LCA({("test", "2"): 1}, method=("m",), presamples=[path])
    lca.lci()
    assert lca.technosphere_matrix[
        lca.product_dict[("test", "1")],
        lca.activity_dict[("test", "2")],
    ] == -1
    lca.presamples.update_matrices()
    assert lca.technosphere_matrix[
        lca.product_dict[("test", "1")],
        lca.activity_dict[("test", "2")],
    ] == -2
    lca.presamples.update_matrices()
    assert lca.technosphere_matrix[
        lca.product_dict[("test", "1")],
        lca.activity_dict[("test", "2")],
    ] == -3
    lca.presamples.update_matrices()
    assert lca.technosphere_matrix[
        lca.product_dict[("test", "1")],
        lca.activity_dict[("test", "2")],
    ] == -1

@pytest.mark.skipif(not PackagesDataLoader, reason="presamples not installed")
def test_sequential_seed_monte_carlo():
    path = os.path.join(basedir, "seq")
    mc = MonteCarloLCA({("test", "2"): 1}, method=("m",), presamples=[path])
    next(mc)
    assert mc.technosphere_matrix[
        mc.product_dict[("test", "1")],
        mc.activity_dict[("test", "2")],
    ] == -1
    next(mc)
    assert mc.technosphere_matrix[
        mc.product_dict[("test", "1")],
        mc.activity_dict[("test", "2")],
    ] == -2
    next(mc)
    assert mc.technosphere_matrix[
        mc.product_dict[("test", "1")],
        mc.activity_dict[("test", "2")],
    ] == -3
    next(mc)
    assert mc.technosphere_matrix[
        mc.product_dict[("test", "1")],
        mc.activity_dict[("test", "2")],
    ] == -1

@pytest.mark.skipif(not PackagesDataLoader, reason="presamples not installed")
def test_call_update_matrices_manually(basic):
    path = os.path.join(basedir, "multi")

    lca = LCA({("test", "2"): 1}, method=("m",), presamples=[path])
    lca.lci()
    lca.lcia()

    results = set()
    for _ in range(100):
        lca.presamples.update_matrices()
        lca.redo_lci()
        lca.redo_lcia()
        results.add(lca.score)

    assert len(results) > 1



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals, division
from eight import *

from scipy import sparse
import numpy as np
from .errors import (
    EmptyBiosphere,
    NonsquareTechnosphere,
    OutsideTechnosphere,
)
from .log_utils import create_logger
from .matrices import MatrixBuilder
from .matrices import TechnosphereBiosphereMatrixBuilder as TBMBuilder
from .utils import (
    global_index,
    clean_databases,
    get_filepaths,
    load_arrays,
    mapping,
    wrap_functional_unit,
)
import copy
import numpy as np
import logging
import warnings

try:
    from pypardiso import factorized, spsolve
except ImportError:
    from scipy.sparse.linalg import factorized, spsolve
try:
    import pandas
except ImportError:
    pandas = None
try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping
try:
    from presamples import PackagesDataLoader
except ImportError:
    PackagesDataLoader = None


class LCA(object):
    """A static LCI or LCIA calculation.

    Following the general philosophy of Brightway2, and good software practices, there is a clear separation of concerns between retrieving and formatting data and doing an LCA. Building the necessary matrices is done with MatrixBuilder objects (:ref:`matrixbuilders`). The LCA class only does the LCA calculations themselves.

    """
    #############
    ### Setup ###
    #############

    def __init__(self, demand, method=None, weighting=None, normalization=None,
                 database_filepath=None, log_config=None, presamples=None,
                 seed=None, override_presamples_seed=False):
        """Create a new LCA calculation.

        Args:
            * *demand* (dict): The demand or functional unit. Needs to be a dictionary to indicate amounts, e.g. ``{("my database", "my process"): 2.5}``.
            * *method* (tuple, optional): LCIA Method tuple, e.g. ``("My", "great", "LCIA", "method")``. Can be omitted if only interested in calculating the life cycle inventory.

        Returns:
            A new LCA object

        """
        if not isinstance(demand, Mapping):
            raise ValueError("Demand must be a dictionary")
        for key in demand:
            if not key:
                raise ValueError("Invalid demand dictionary")

        if log_config:
            create_logger(**log_config)
        self.logger = logging.getLogger('bw2calc')

        clean_databases()
        self._fixed = False

        self.demand = demand
        self.method = method
        self.normalization = normalization
        self.weighting = weighting
        self.database_filepath = database_filepath
        self.seed = seed

        if presamples and PackagesDataLoader is None:
            warnings.warn("Skipping presamples; `presamples` not installed")
            self.presamples = None
        elif presamples:
            # Iterating over a `Campaign` object will also return the presample filepaths
            self.presamples = PackagesDataLoader(
                dirpaths=presamples,
                seed=self.seed if override_presamples_seed else None,
                lca=self
            )
        else:
            self.presamples = None

        self.database_filepath, \
            self.method_filepath, \
            self.weighting_filepath, \
            self.normalization_filepath = \
            self.get_array_filepaths()

        self.logger.info("Created LCA object", extra={
            'demand': wrap_functional_unit(self.demand),
            'database_filepath': self.database_filepath,
            'method': self.method,
            'method_filepath': self.method_filepath,
            'normalization': self.normalization,
            'normalization_filepath': self.normalization_filepath,
            'presamples': str(self.presamples),
            'weighting': self.weighting,
            'weighting_filepath': self.weighting_filepath,
        })

    def get_array_filepaths(self):
        """Use utility functions to get all array filepaths"""
        return (
            get_filepaths(self.demand, "demand"),
            get_filepaths(self.method, "method"),
            get_filepaths(self.weighting, "weighting"),
            get_filepaths(self.normalization, "normalization"),
        )

    def build_demand_array(self, demand=None):
        """Turn the demand dictionary into a *NumPy* array of correct size.

        Args:
            * *demand* (dict, optional): Demand dictionary. Optional, defaults to ``self.demand``.

        Returns:
            A 1-dimensional NumPy array

        """
        demand = demand or self.demand
        self.demand_array = np.zeros(len(self.product_dict))
        for key in demand:
            try:
                self.demand_array[self.product_dict[key]] = demand[key]
            except KeyError:
                if key in self.activity_dict:
                    raise ValueError((u"LCA can only be performed on products,"
                        u" not activities ({} is the wrong dimension)"
                        ).format(key)
                    )
                else:
                    raise OutsideTechnosphere("Can't find key {} in product dictionary".format(key))

    #########################
    ### Data manipulation ###
    #########################

    def fix_dictionaries(self):
        """
Fix technosphere and biosphere dictionaries from this:

.. code-block:: python

    {mapping integer id: matrix row/column index}

To this:

.. code-block:: python

    {(database, key): matrix row/column index}

This isn't needed for the LCA calculation itself, but is helpful when interpreting results.

Doesn't require any arguments or return anything, but changes ``self.activity_dict``, ``self.product_dict`` and ``self.biosphere_dict``.

        """
        if self._fixed:
            # Already fixed - should be idempotent
            return False
        elif not mapping:
            # Don't have access to mapping
            return False
        rev_mapping = {v: k for k, v in mapping.items()}
        self._activity_dict = copy.deepcopy(self.activity_dict)
        self.activity_dict = {
            rev_mapping[k]: v for k, v in self.activity_dict.items()}
        self._product_dict = self.product_dict
        self.product_dict = {
            rev_mapping[k]: v for k, v in self.product_dict.items()}
        self._biosphere_dict = self.biosphere_dict
        self.biosphere_dict = {
            rev_mapping[k]: v for k, v in self.biosphere_dict.items()}
        self._fixed = True
        return True

    def reverse_dict(self):
        """Construct reverse dicts from technosphere and biosphere row and col indices to activity_dict/product_dict/biosphere_dict keys.

        Returns:
            (reversed ``self.activity_dict``, ``self.product_dict`` and ``self.biosphere_dict``)
        """
        rev_activity = {v: k for k, v in self.activity_dict.items()}
        rev_product = {v: k for k, v in self.product_dict.items()}
        rev_bio = {v: k for k, v in self.biosphere_dict.items()}
        return rev_activity, rev_product, rev_bio

    ######################
    ### Data retrieval ###
    ######################

    def load_lci_data(self, fix_dictionaries=True, builder=TBMBuilder):
        """Load data and create technosphere and biosphere matrices."""
        self._fixed = False
        self.bio_params, self.tech_params, \
            self.biosphere_dict, self.activity_dict, \
            self.product_dict, self.biosphere_matrix, \
            self.technosphere_matrix = \
            builder.build(self.database_filepath)
        if len(self.activity_dict) != len(self.product_dict):
            raise NonsquareTechnosphere((
                "Technosphere matrix is not square: {} activities (columns) and {} products (rows). "
                "Use LeastSquaresLCA to solve this system, or fix the input "
                "data").format(len(self.activity_dict), len(self.product_dict))
            )
        if fix_dictionaries:
            self.fix_dictionaries()

        if not self.biosphere_dict:
            warnings.warn("No biosphere flows found. No inventory results can "
                          "be calculated, `lcia` will raise an error")

        # Only need to index here for traditional LCA
        if self.presamples:
            self.presamples.index_arrays(self)
            self.presamples.update_matrices(
                matrices=('technosphere_matrix', 'biosphere_matrix')
            )

    def load_lcia_data(self, builder=MatrixBuilder):
        """Load data and create characterization matrix.

        This method will filter out regionalized characterization factors. This filtering needs access to ``bw2data`` - therefore, regionalized methods will cause incorrect results if ``bw2data`` is not importable.

        """
        self.cf_params, _, _, self.characterization_matrix = builder.build(
                self.method_filepath,
                "amount",
                "flow",
                "row",
                row_dict=self._biosphere_dict,
                one_d=True,
            )
        if global_index is not None:
            mask = self.cf_params['geo'] == global_index
            self.cf_params = self.cf_params[mask]
            self.characterization_matrix = builder.build_diagonal_matrix(self.cf_params, self._biosphere_dict, "row", "amount")

        if self.presamples:
            self.presamples.update_matrices(matrices=['characterization_matrix'])

    def load_normalization_data(self, builder=MatrixBuilder):
        """Load normalization data."""
        self.normalization_params, _, _, self.normalization_matrix = \
            builder.build(
                self.normalization_filepath,
                "amount",
                "flow",
                "index",
                row_dict=self._biosphere_dict,
                one_d=True
            )
        if self.presamples:
            self.presamples.update_matrices(matrices=['normalization_matrix',])

    def load_weighting_data(self):
        """Load weighting data, a 1-element array."""
        self.weighting_params = load_arrays(
            self.weighting_filepath
        )
        self.weighting_value = self.weighting_params['amount']

        # TODO: This won't work because weighting is a value not a matrix
        # if self.presamples:
        #     self.presamples.update_matrices(self, ['weighting_value',])

    ####################
    ### Calculations ###
    ####################

    def decompose_technosphere(self):
        """
Factorize the technosphere matrix into lower and upper triangular matrices, :math:`A=LU`. Does not solve the linear system :math:`Ax=B`.

Doesn't return anything, but creates ``self.solver``.

.. warning:: Incorrect results could occur if a technosphere matrix was factorized, and then a new technosphere matrix was constructed, as ``self.solver`` would still be the factorized older technosphere matrix. You are responsible for deleting ``self.solver`` when doing these types of advanced calculations.

        """
        self.solver = factorized(self.technosphere_matrix.tocsc())

    def solve_linear_system(self):
        """
Master solution function for linear system :math:`Ax=B`.

    To most numerical analysts, matrix inversion is a sin.

    -- Nicolas Higham, Accuracy and Stability of Numerical Algorithms, Society for Industrial and Applied Mathematics, Philadelphia, PA, USA, 2002, p. 260.

We use `UMFpack <http://www.cise.ufl.edu/research/sparse/umfpack/>`_, which is a very fast solver for sparse matrices.

If the technosphere matrix has already been factorized, then the decomposed technosphere (``self.solver``) is reused. Otherwise the calculation is redone completely.

        """
        if hasattr(self, "solver"):
            return self.solver(self.demand_array)
        else:
            return spsolve(
                self.technosphere_matrix,
                self.demand_array)

    def lci(self, factorize=False,
            builder=TBMBuilder):
        """
Calculate a life cycle inventory.

#. Load LCI data, and construct the technosphere and biosphere matrices.
#. Build the demand array
#. Solve the linear system to get the supply array and life cycle inventory.

Args:
    * *factorize* (bool, optional): Factorize the technosphere matrix. Makes additional calculations with the same technosphere matrix much faster. Default is ``False``; not useful is only doing one LCI calculation.
    * *builder* (``MatrixBuilder`` object, optional): Default is ``bw2calc.matrices.TechnosphereBiosphereMatrixBuilder``, which is fine for most cases. Custom matrix builders can be used to manipulate data in creative ways before building the matrices.

.. warning:: Custom matrix builders should inherit from ``TechnosphereBiosphereMatrixBuilder``, because technosphere inputs need to have their signs flipped to be negative, as we do :math:`A^{-1}f` directly instead of :math:`(I - A^{-1})f`.

Doesn't return anything, but creates ``self.supply_array`` and ``self.inventory``.

        """
        self.load_lci_data(builder=builder)
        self.build_demand_array()
        if factorize:
            self.decompose_technosphere()
        self.lci_calculation()

    def lci_calculation(self):
        """The actual LCI calculation.

        Separated from ``lci`` to be reusable in cases where the matrices are already built, e.g. ``redo_lci`` and Monte Carlo classes.

        """
        self.supply_array = self.solve_linear_system()
        # Turn 1-d array into diagonal matrix
        count = len(self.activity_dict)
        self.inventory = self.biosphere_matrix * \
            sparse.spdiags([self.supply_array], [0], count, count)

    def lcia(self, builder=MatrixBuilder):
        """
Calculate the life cycle impact assessment.

#. Load and construct the characterization matrix
#. Multiply the characterization matrix by the life cycle inventory

Args:
    * *builder* (``MatrixBuilder`` object, optional): Default is ``bw2calc.matrices.MatrixBuilder``, which is fine for most cases. Custom matrix builders can be used to manipulate data in creative ways before building the characterization matrix.

Doesn't return anything, but creates ``self.characterized_inventory``.

        """
        assert hasattr(self, "inventory"), "Must do lci first"
        assert self.method, "Must specify a method to perform LCIA"
        if not self.biosphere_dict:
            raise EmptyBiosphere

        self.load_lcia_data(builder)
        self.lcia_calculation()

    def lcia_calculation(self):
        """The actual LCIA calculation.

        Separated from ``lcia`` to be reusable in cases where the matrices are already built, e.g. ``redo_lcia`` and Monte Carlo classes.

        """
        self.characterized_inventory = \
            self.characterization_matrix * self.inventory

    def normalize(self):
        """Multiply characterized inventory by flow-specific normalization factors."""
        assert hasattr(self, "characterized_inventory"), "Must do lcia first"
        if not hasattr(self, "normalization_matrix"):
            self.load_normalization_data()
        self.normalization_calculation()

    def normalization_calculation(self):
        """The actual normalization calculation.

        Creates ``self.normalized_inventory``."""
        self.normalized_inventory = \
            self.normalization_matrix * self.characterized_inventory

    def weight(self):
        """Multiply characterized inventory by weighting value.

        Can be done with or without normalization."""
        assert hasattr(self, "characterized_inventory"), "Must do lcia first"
        if not hasattr(self, "weighting_value"):
            self.load_weighting_data()

    def weighting_calculation(self):
        """The actual weighting calculation.

        Multiples weighting value by normalized inventory, if available, otherwise by characterized inventory.

        Creates ``self.weighted_inventory``."""
        if hasattr(self, "normalized_inventory"):
            obj = self.normalized_inventory
        else:
            obj = self.characterized_inventory
        self.weighted_inventory = self.weighting_value[0] * obj

    @property
    def score(self):
        """
The LCIA score as a ``float``.

Note that this is a `property <http://docs.python.org/2/library/functions.html#property>`_, so it is ``foo.lca``, not ``foo.score()``
        """
        assert hasattr(self, "characterized_inventory"), "Must do LCIA first"
        if self.weighting:
            assert hasattr(self, "weighted_inventory"), "Must do weighting first"
            return float(self.weighted_inventory.sum())
        return float(self.characterized_inventory.sum())

    #########################
    ### Redo calculations ###
    #########################

    def rebuild_technosphere_matrix(self, vector):
        """Build a new technosphere matrix using the same row and column indices, but different values. Useful for Monte Carlo iteration or sensitivity analysis.

        Args:
            * *vector* (array): 1-dimensional NumPy array with length (# of technosphere parameters), in same order as ``self.tech_params``.

        Doesn't return anything, but overwrites ``self.technosphere_matrix``.

        """
        self.technosphere_matrix = MatrixBuilder.build_matrix(
            self.tech_params, self._activity_dict, self._product_dict,
            "row", "col",
            new_data=TBMBuilder.fix_supply_use(self.tech_params, vector.copy())
        )

    def rebuild_biosphere_matrix(self, vector):
        """Build a new biosphere matrix using the same row and column indices, but different values. Useful for Monte Carlo iteration or sensitivity analysis.

        Args:
            * *vector* (array): 1-dimensional NumPy array with length (# of biosphere parameters), in same order as ``self.bio_params``.

        Doesn't return anything, but overwrites ``self.biosphere_matrix``.

        """
        self.biosphere_matrix = MatrixBuilder.build_matrix(
            self.bio_params, self._biosphere_dict, self._activity_dict,
            "row", "col", new_data=vector)

    def rebuild_characterization_matrix(self, vector):
        """Build a new characterization matrix using the same row and column indices, but different values. Useful for Monte Carlo iteration or sensitivity analysis.

        Args:
            * *vector* (array): 1-dimensional NumPy array with length (# of characterization parameters), in same order as ``self.cf_params``.

        Doesn't return anything, but overwrites ``self.characterization_matrix``.

        """
        self.characterization_matrix = MatrixBuilder.build_diagonal_matrix(
            self.cf_params, self._biosphere_dict,
            "row", "row", new_data=vector)

    def switch_method(self, method):
        """Switch to LCIA method `method`"""
        self.method = method
        _, self.method_filepath, _, _ = self.get_array_filepaths()
        self.load_lcia_data()
        self.logger.info("Switching LCIA method", extra={'method': method})

    def switch_normalization(self, normalization):
        """Switch to LCIA normalization `normalization`"""
        self.normalization = normalization
        _, _, _, self.normalization_filepath = self.get_array_filepaths()
        self.load_normalization_data()
        self.logger.info("Switching LCIA normalization",
                         extra={'normalization': normalization})

    def switch_weighting(self, weighting):
        """Switch to LCIA weighting `weighting`"""
        self.weighting = weighting
        _, _, self.weighting_filepath, _ = self.get_array_filepaths()
        self.load_weighting_data()
        self.logger.info("Switching LCIA weighting", extra={'weighting': weighting})

    def redo_lci(self, demand=None):
        """Redo LCI with same databases but different demand.

        Args:
            * *demand* (dict): A demand dictionary.

        Doesn't return anything, but overwrites ``self.demand_array``, ``self.supply_array``, and ``self.inventory``.

        .. warning:: If you want to redo the LCIA as well, use ``redo_lcia(demand)`` directly.

        """
        assert hasattr(self, "inventory"), "Must do lci first"
        if demand is not None:
            self.build_demand_array(demand)
        self.lci_calculation()
        self.demand = demand
        self.logger.info("Redoing LCI", extra={'demand': wrap_functional_unit(demand or self.demand)})

    def redo_lcia(self, demand=None):
        """Redo LCIA, optionally with new demand.

        Args:
            * *demand* (dict, optional): New demand dictionary. Optional, defaults to ``self.demand``.

        Doesn't return anything, but overwrites ``self.characterized_inventory``. If ``demand`` is given, also overwrites ``self.demand_array``, ``self.supply_array``, and ``self.inventory``.

        """
        assert hasattr(self, "characterized_inventory"), "Must do LCIA first"
        if demand is not None:
            self.redo_lci(demand)
        self.lcia_calculation()
        self.demand = demand
        self.logger.info("Redoing LCIA", extra={'demand': wrap_functional_unit(demand or self.demand)})

    def to_dataframe(self, cutoff=200):
        """Return all nonzero elements of characterized inventory as Pandas dataframe"""
        assert mapping, "This method doesn't work with independent LCAs"
        assert pandas, "This method requires the `pandas` (http://pandas.pydata.org/) library"
        assert hasattr(self, "characterized_inventory"), "Must do LCIA calculation first"

        from bw2data import get_activity

        coo = self.characterized_inventory.tocoo()
        stacked = np.vstack([np.abs(coo.data), coo.row, coo.col, coo.data])
        stacked.sort()
        rev_activity, _, rev_bio = self.reverse_dict()
        length = stacked.shape[1]

        data = []
        for x in range(min(cutoff, length)):
            if stacked[3, length - x - 1] == 0.:
                continue
            activity = get_activity(rev_activity[stacked[2, length - x - 1]])
            flow = get_activity(rev_bio[stacked[1, length - x - 1]])
            data.append((
                activity['name'],
                flow['name'],
                activity.get('location'),
                stacked[3, length - x - 1]
            ))
        return pandas.DataFrame(
            data,
            columns=['Activity', 'Flow', 'Region', 'Amount']
        )

    ####################
    ### Contribution ###
    ####################

    def top_emissions(self, **kwargs):
        """Call ``bw2analyzer.ContributionAnalyses.annotated_top_emissions``"""
        try:
            from bw2analyzer import ContributionAnalysis
        except ImportError:
            raise ImportError("`bw2analyzer` is not installed")
        return ContributionAnalysis().annotated_top_emissions(self, **kwargs)

    def top_activities(self, **kwargs):
        """Call ``bw2analyzer.ContributionAnalyses.annotated_top_processes``"""
        try:
            from bw2analyzer import ContributionAnalysis
        except ImportError:
            raise ImportError("`bw2analyzer` is not installed")
        return ContributionAnalysis().annotated_top_processes(self, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
from eight import *

from bw2calc import *
from bw2data import config, Database
from bw2data.tests import BW2DataTest
from bw2data.utils import MAX_INT_32, numpy_string
import numpy as np

TBM = TechnosphereBiosphereMatrixBuilder


class MatrixBuilderTestCase(BW2DataTest):
    def test_build_one_d(self):
        database = Database("sour")
        database.register()
        dtype = [
            (numpy_string('a'), np.uint32),
            (numpy_string('row'), np.uint32),
            (numpy_string('values'), np.float32),
        ]
        array = np.array([
            (1, MAX_INT_32, 99),
            (2, MAX_INT_32, 100),
            ], dtype=dtype
        )
        row_dict = {1: 0, 2: 1}
        np.save(database.filepath_processed(), array, allow_pickle=False)
        matrix = MatrixBuilder.build([database.filepath_processed()],
            "values", "a", "row",
            row_dict=row_dict, one_d=True)[3]
        self.assertTrue(np.allclose(
            matrix.todense(),
            np.array(((99, 0), (0, 100)))
        ))

    def test_build_one_d_drop_missing(self):
        database = Database("ghost")
        database.register()
        dtype = [
            (numpy_string('a'), np.uint32),
            (numpy_string('row'), np.uint32),
            (numpy_string('values'), np.float32),
        ]
        array = np.array([
            (1, MAX_INT_32, 99),
            (2, MAX_INT_32, 99),
            (3, MAX_INT_32, 99),
            ], dtype=dtype
        )
        row_dict = {1: 0, 2: 1}
        np.save(database.filepath_processed(), array, allow_pickle=False)
        values = MatrixBuilder.build([database.filepath_processed()],
            "values", "a", "row",
            row_dict=row_dict, one_d=True)[0]
        self.assertEqual(values.shape, (2,))

    def test_one_d_missing_in_row_dict_raise_valueerror(self):
        database = Database("ghost")
        database.register()
        dtype = [
            (numpy_string('a'), np.uint32),
            (numpy_string('row'), np.uint32),
            (numpy_string('values'), np.float32),
        ]
        array = np.array([
            (1, MAX_INT_32, 99),
            (2, MAX_INT_32, 99),
            ], dtype=dtype
        )
        row_dict = {1: 0}
        np.save(database.filepath_processed(), array, allow_pickle=False)
        with self.assertRaises(ValueError):
            MatrixBuilder.build([database.filepath_processed()],
                "values", "a", "row",
                row_dict=row_dict, one_d=True, drop_missing=False)

    def test_build_drop_missing(self):
        database = Database("boo")
        database.register()
        dtype = [
            (numpy_string('a'), np.uint32),
            (numpy_string('b'), np.uint32),
            (numpy_string('row'), np.uint32),
            (numpy_string('col'), np.uint32),
            (numpy_string('values'), np.float32),
        ]
        array = np.array([
            (1, 2, MAX_INT_32, MAX_INT_32, 99),
            (3, 4, MAX_INT_32, MAX_INT_32, 99),
            (3, 2, MAX_INT_32, MAX_INT_32, 99),
            (5, 6, MAX_INT_32, MAX_INT_32, 99),
            ], dtype=dtype
        )
        row_dict = {1: 0, 3: 1}
        col_dict = {2: 0, 6: 1}
        np.save(database.filepath_processed(), array, allow_pickle=False)
        values = MatrixBuilder.build([database.filepath_processed()],
            "values", "a", "row", "b", "col", row_dict, col_dict)[0]
        self.assertEqual(values.shape, (2,))

    def test_missing_in_row_dict_raise_valueerror(self):
        database = Database("whoah")
        database.register()
        dtype = [
            (numpy_string('a'), np.uint32),
            (numpy_string('b'), np.uint32),
            (numpy_string('row'), np.uint32),
            (numpy_string('col'), np.uint32),
            (numpy_string('values'), np.float32),
        ]
        array = np.array([
            (1, 2, MAX_INT_32, MAX_INT_32, 99),
            (1, 4, MAX_INT_32, MAX_INT_32, 99),
            ], dtype=dtype
        )
        row_dict = {1: 0}
        col_dict = {2: 0}
        np.save(database.filepath_processed(), array, allow_pickle=False)
        with self.assertRaises(ValueError):
            MatrixBuilder.build([database.filepath_processed()],
                "values", "a", "row", "b", "col", row_dict, col_dict,
                drop_missing=False)

    def test_build_matrix(self):
        a = np.zeros((4,), dtype=[
            (numpy_string('values'), np.float64),
            (numpy_string('rows'), np.uint32),
            (numpy_string('cols'), np.uint32)
        ])
        a[0] = (4.2, 0, 2)
        a[1] = (6.6, 1, 1)
        a[2] = (1.3, 2, 1)
        a[3] = (10, 2, 2)
        r = [0,0,0]  # Just need right length
        matrix = MatrixBuilder.build_matrix(
            array=a,
            row_dict=r,
            col_dict=r,
            row_index_label='rows',
            col_index_label='cols',
            data_label='values'
        )
        answer = np.array((
            (0, 0, 4.2),
            (0, 6.6, 0),
            (0, 1.3, 10)
        ))
        self.assertTrue(np.allclose(answer, matrix.todense()))

    def test_multiple_values_same_exchange(self):
        """Values for same (row, col) should add together"""
        a = np.zeros((2,), dtype=[
            (numpy_string('values'), np.float64),
            (numpy_string('rows'), np.uint32),
            (numpy_string('cols'), np.uint32)
        ])
        a[0] = (9, 1, 1)
        a[1] = (33, 1, 1)
        r = [0,0,0]  # Just need right length
        matrix = MatrixBuilder.build_matrix(
            array=a,
            row_dict=r,
            col_dict=r,
            row_index_label='rows',
            col_index_label='cols',
            data_label='values'
        )
        answer = np.array((
            (0, 0, 0),
            (0, 42, 0),
            (0, 0, 0)
        ))
        self.assertTrue(np.allclose(answer, matrix.todense()))



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
from eight import *

from bw2calc.single_matrix import SingleMatrixLCA
import numpy as np
import pytest
import os


fixture = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "fixtures",
    "single-matrix",
    "sm-fixture.tar.bz2"
)


def test_single_matrix():
    s = SingleMatrixLCA({"f": 1}, "sm-fixture.tar.bz2")
    s.calculate()
    assert s.scores['foo']



from bw2calc import LCA, IndependentLCAMixin
from bw2data.tests import BW2DataTest, bw2test
from bw2data.utils import TYPE_DICTIONARY, MAX_INT_32
from io import BytesIO
import numpy as np
import os

basedir = os.path.join(os.path.dirname(__file__), "fixtures", "independent")
inv_fp = os.path.join(basedir, "inv.npy")
ia_fp = os.path.join(basedir, "ia.npy")


@bw2test
def test_independent_lca_with_global_value(monkeypatch):
    monkeypatch.setattr(
        'bw2calc.lca.global_index',
        17
    )

    class ILCA(IndependentLCAMixin, LCA):
        pass

    lca = ILCA({15: 1}, database_filepath=[inv_fp], method=[ia_fp])
    lca.lci()
    lca.lcia()
    print(lca.score)
    assert lca.score == 8020


@bw2test
def test_independent_lca_with_no_global_value(monkeypatch):
    monkeypatch.setattr(
        'bw2calc.lca.global_index',
        None
    )

    class ILCA(IndependentLCAMixin, LCA):
        pass

    lca = ILCA({15: 1}, database_filepath=[inv_fp], method=[ia_fp])
    lca.lci()
    lca.lcia()
    print(lca.score)
    assert lca.score == 8020


@bw2test
def test_independent_lca_with_directly_passing_array(monkeypatch):
    monkeypatch.setattr(
        'bw2calc.lca.global_index',
        None
    )

    class ILCA(IndependentLCAMixin, LCA):
        pass

    ia = np.load(ia_fp, allow_pickle=True)
    lca = ILCA({15: 1}, database_filepath=[inv_fp], method=[ia])
    lca.lci()
    lca.lcia()
    assert lca.score == 8020


@bw2test
def test_independent_lca_with_passing_bytes_array(monkeypatch):
    monkeypatch.setattr(
        'bw2calc.lca.global_index',
        None
    )

    class ILCA(IndependentLCAMixin, LCA):
        pass

    with BytesIO() as buffer:
        np.save(buffer, np.load(ia_fp, allow_pickle=True))
        buffer.seek(0)
        lca = ILCA({15: 1}, database_filepath=[inv_fp], method=[buffer])
        lca.lci()
        lca.lcia()
        assert lca.score == 8020


if __name__ == '__main__':

    """

    Biosphere flows:

    10: A
    11: B

    Activities:

    12: C
    13: D
    14: E

    Products:

    15: F
    16: G
    17: H

    CFs:

        A: 1, B: 10

    Exchanges:

        F -> C: 1, G -> D: 1, H -> E: 1
        G -> C: 2
        H -> D: 4
        A -> D: 10
        B -> E: 100

    """
    inv_data = [
        (15, 12, 1,   'production'),
        (16, 13, 1,   'production'),
        (17, 14, 1,   'production'),
        (16, 12, 2,   'technosphere'),
        (17, 13, 4,   'technosphere'),
        (10, 13, 10,  'biosphere'),
        (11, 14, 100, 'biosphere'),
    ]

    dtype = [
        ('input', np.uint32),
        ('output', np.uint32),
        ('row', np.uint32),
        ('col', np.uint32),
        ('type', np.uint8),
        ('amount', np.float32),
    ]
    arr = np.zeros((len(inv_data),), dtype=dtype)

    for index, line in enumerate(inv_data):
        arr[index] = (
            line[0],
            line[1],
            MAX_INT_32,
            MAX_INT_32,
            TYPE_DICTIONARY[line[3]],
            line[2],
        )

    arr.sort(order=['input', 'output', 'amount'])
    np.save(inv_fp, arr, allow_pickle=False)

    ia_data = [
        (10, 1),
        (11, 10),
    ]

    dtype = [
        ('flow', np.uint32),
        ('row', np.uint32),
        ('col', np.uint32),
        ('geo', np.uint32),
        ('amount', np.float32),
    ]
    arr = np.zeros((len(ia_data),), dtype=dtype)

    for index, line in enumerate(ia_data):
        arr[index] = (
            line[0],
            MAX_INT_32,
            MAX_INT_32,
            17,
            line[1],
        )

    arr.sort(order=['flow', 'amount'])
    np.save(ia_fp, arr, allow_pickle=False)



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
from eight import *

class BW2CalcError(Exception):
    """Base class for bw2calc errors"""
    pass

class OutsideTechnosphere(BW2CalcError):
    """The given demand array activity is not in the technosphere matrix"""
    pass


class EfficiencyWarning(RuntimeWarning):
    """Least squares is much less efficient than direct computation for square, full-rank matrices"""
    pass


class NoSolutionFound(UserWarning):
    """No solution to set of linear equations found within given constraints"""
    pass


class NonsquareTechnosphere(BW2CalcError):
    """The given data do not form a square technosphere matrix"""
    pass


class MalformedFunctionalUnit(BW2CalcError):
    """The given functional unit cannot be understood"""
    pass


class EmptyBiosphere(BW2CalcError):
    """Can't do impact assessment with no biosphere flows"""
    pass


class AllArraysEmpty(BW2CalcError):
    """Can't load the numpy arrays if all of them are empty"""
    pass



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
from eight import *

import numpy as np
from scipy.sparse.linalg import spsolve
from scipy.sparse import coo_matrix
from time import time


class SpeedTest(object):
    """Compare speed of sparse matrix operations on this machine compared to a reference machine."""
    size = 1000
    density = 0.02
    seed = 42

    def test(self):
        sm = self.get_sparse_matrix()
        v = self.get_demand_vector()
        now = time()
        for x in range(100):
            spsolve(sm, v)
        return time() - now

    def ratio(self):
        """On the reference machine, this takes about 5.85 seconds"""
        return 5.85 / self.test()

    def get_sparse_matrix(self):
        """Adapted from scipy to use seeded RNG"""
        self.size = 1000
        k = long(self.density * self.size ** 2)
        rng = np.random.RandomState(self.seed)
        i = rng.randint(self.size, size=k)
        j = rng.randint(self.size, size=k)
        data = rng.rand(k)
        return coo_matrix((data, (i, j)), shape=(self.size, self.size)).tocsr()

    def get_demand_vector(self):
        v = np.zeros((self.size,))
        v[42] = 1
        return v



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals, division
from eight import *

from scipy import sparse
from io import BytesIO
import numpy as np
from .errors import (
    NonsquareTechnosphere,
    OutsideTechnosphere,
)
from .log_utils import create_logger
from .matrices import SingleMatrixBuilder, MatrixBuilder
from .utils import (
    global_index,
    clean_databases,
    get_filepaths,
    load_arrays,
    mapping,
)
import copy
import numpy as np
import json
import logging
import tarfile
import warnings

try:
    from pypardiso import factorized, spsolve
except ImportError:
    from scipy.sparse.linalg import factorized, spsolve
try:
    import pandas
except ImportError:
    pandas = None
try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping
try:
    from presamples import PackagesDataLoader
except ImportError:
    PackagesDataLoader = None


class SingleMatrixLCA(object):
    """An LCA which puts everything into one matrix.

    Comes with advantages and disadvantages, and designed exclusively for `BONSAI <https://bonsai.uno/>`__ via `beebee <https://github.com/BONSAMURAIS/beebee/>`__."""
    def __init__(self, demand, data_filepath, log_config=None, presamples=None,
                 seed=None, override_presamples_seed=False):
        """Create a new single-matrix LCA calculation.

        Args:
            * *demand* (dict): The demand or functional unit. Needs to be a dictionary to indicate amounts, e.g. ``{("my database", "my process"): 2.5}``.

        Returns:
            A new ``SingleMatrixLCA`` object

        """
        if not isinstance(demand, Mapping):
            raise ValueError("Demand must be a dictionary")

        if log_config:
            create_logger(**log_config)
        self.logger = logging.getLogger('bw2calc')

        self.demand = demand
        self.filepath = data_filepath
        self.seed = seed

        if presamples and PackagesDataLoader is None:
            warnings.warn("Skipping presamples; `presamples` not installed")
            self.presamples = None
        elif presamples:
            # Iterating over a `Campaign` object will also return the presample filepaths
            self.presamples = PackagesDataLoader(
                dirpaths=presamples,
                seed=self.seed if override_presamples_seed else None,
                lca=self
            )
        else:
            self.presamples = None

        self.logger.info("Created LCA object", extra={
            'demand': str(self.demand),
            'data_filepath': self.filepath,
            'presamples': str(self.presamples),
        })

    def build_demand_array(self, demand=None):
        """Turn the demand dictionary into a *NumPy* array of correct size.

        Args:
            * *demand* (dict, optional): Demand dictionary. Optional, defaults to ``self.demand``.

        Returns:
            A 1-dimensional NumPy array

        """
        demand = demand or self.demand
        self.demand_array = np.zeros(len(self.row_dict))
        for key in demand:
            try:
                self.demand_array[self.row_dict[key]] = demand[key]
            except KeyError:
                if key in self.col_dict:
                    raise ValueError((u"LCA can only be performed on products,"
                        u" not activities ({} is the wrong dimension)"
                        ).format(key)
                    )
                else:
                    raise OutsideTechnosphere("Can't find key {} in product dictionary".format(key))

    #########################
    ### Data manipulation ###
    #########################

    def fix_dictionaries(self, row_mapping, col_mapping):
        """Fix the row and column dictionaries from ``{integer: row/col index}`` to ``{label: row/col index}``."""
        self.row_dict = {label: self.row_dict[index] for label, index in row_mapping.items()}
        self.col_dict = {label: self.col_dict[index] for label, index in col_mapping.items()}

    def reverse_dict(self):
        """Construct reverse dicts from technosphere and biosphere row and col indices to activity_dict/product_dict/biosphere_dict keys.

        Returns:
            (reversed ``self.activity_dict``, ``self.product_dict`` and ``self.biosphere_dict``)
        """
        rev_row = {v: k for k, v in self.row_dict.items()}
        rev_col = {v: k for k, v in self.col_dict.items()}
        return rev_row, rev_col

    ######################
    ### Data retrieval ###
    ######################

    def load_beebee_data(self, builder=SingleMatrixBuilder):
        """Load ``beebee`` export data package.

        This is a compressed file which contains:

        * A `stats_arrays <https://bitbucket.org/cmutel/stats_arrays>`__ file used to create the single matrix.
        * A mapping dictionary from meaningful labels to the integer row ids
        * A mapping dictionary from meaningful labels to the integer column ids
        * A mapping dictionary from ``{"method URI": {labels}}`` which allows for LCIA sums

        """
        with tarfile.open(self.filepath, 'r:bz2') as f:

            # Hack needed because of https://github.com/numpy/numpy/issues/7989
            array_file = BytesIO()
            array_file.write(f.extractfile("array.npy").read())
            array_file.seek(0)

            self.params, self.row_dict, self.col_dict, \
                self.matrix = builder.build(array_file)
            row_mapping = json.load(f.extractfile("row.mapping"))
            col_mapping = json.load(f.extractfile("col.mapping"))
            categories = json.load(f.extractfile("categories.mapping"))
        if len(self.row_dict) != len(self.col_dict):
            raise NonsquareTechnosphere((
                "Single matrix is not square: {} columns and {} rows."
                ).format(len(self.row_dict), len(self.col_dict))
            )
        self.fix_dictionaries(row_mapping, col_mapping)

        self.category_index_arrays = {
            name: np.array(
                [(self.row_dict[label], 1) for label in labels],
                dtype=[('INDEX', np.uint32), ('CONSTANT', np.uint32)]
            ) for name, labels in categories.items()
        }

        # Only need to index here for traditional LCA
        if self.presamples:
            self.presamples.index_arrays(self)
            self.presamples.update_matrices(
                matrices=('matrix',)
            )

    def decompose_technosphere(self):
        """
Factorize the technosphere matrix into lower and upper triangular matrices, :math:`A=LU`. Does not solve the linear system :math:`Ax=B`.

Doesn't return anything, but creates ``self.solver``.

.. warning:: Incorrect results could occur if a technosphere matrix was factorized, and then a new technosphere matrix was constructed, as ``self.solver`` would still be the factorized older technosphere matrix. You are responsible for deleting ``self.solver`` when doing these types of advanced calculations.

        """
        self.solver = factorized(self.matrix.tocsc())

    def solve_linear_system(self):
        """
Master solution function for linear system :math:`Ax=B`.

    To most numerical analysts, matrix inversion is a sin.

    -- Nicolas Higham, Accuracy and Stability of Numerical Algorithms, Society for Industrial and Applied Mathematics, Philadelphia, PA, USA, 2002, p. 260.

We use `UMFpack <http://www.cise.ufl.edu/research/sparse/umfpack/>`_, which is a very fast solver for sparse matrices.

If the technosphere matrix has already been factorized, then the decomposed technosphere (``self.solver``) is reused. Otherwise the calculation is redone completely.

        """
        if hasattr(self, "solver"):
            return self.solver(self.demand_array)
        else:
            return spsolve(
                self.matrix,
                self.demand_array)

    def lcia(self, *args, **kwargs):
        warnings.warn("LCIA does nothing in a SingleMatrixLCA")

    def calculate(self, factorize=False,
            builder=SingleMatrixBuilder):
        """Calculate an LCA score.

        Creates ``self.supply_array``, a vector of activities, flows, and characterization pathways which satisfy the demand.

        Creates ``self.scores``, a dictionary of ``{'LCIA identifier': LCA score}``.

        Create ``self.contributions``, a dictionary of ``{'LCIA identifier': []}``.

        Args:
            * *factorize* (bool, optional): Factorize the technosphere matrix. Makes additional calculations with the same technosphere matrix much faster. Default is ``False``; not useful is only doing one LCI calculation.
            * *builder* (``SingleMatrixBuilder`` object, optional): Custom matrix builders can be used to manipulate data in creative ways before building the matrices.

        Doesn't return anything.

        """
        self.load_beebee_data(builder)
        self.build_demand_array()
        if factorize:
            self.decompose_technosphere()
        self.calculate_scores()

    def calculate_scores(self):
        self.supply_array = self.solve_linear_system()
        self.scores, self.contributions = {}, {}
        for name, array in self.category_index_arrays.items():
            matrix = MatrixBuilder.build_diagonal_matrix(
                array, self.row_dict, 'INDEX', data_label='CONSTANT'
            )
            self.scores[name] = float((self.matrix * matrix * self.supply_array).sum())

    def rebuild_matrix(self, vector):
        """Build a new technosphere matrix using the same row and column indices, but different values. Useful for Monte Carlo iteration or sensitivity analysis.

        Args:
            * *vector* (array): 1-dimensional NumPy array with length (# of technosphere parameters), in same order as ``self.tech_params``.

        Doesn't return anything, but overwrites ``self.technosphere_matrix``.

        """
        self.matrix = SingleMatrixBuilder.build_single_matrix(
            self.params, self.row_dict, self.col_dict, vector
        )

    def redo_calculate(self, demand=None):
        """Redo LCI with same databases but different demand.

        Args:
            * *demand* (dict): A demand dictionary.

        Doesn't return anything, but overwrites ``self.demand_array``, ``self.supply_array``, and ``self.inventory``.

        .. warning:: If you want to redo the LCIA as well, use ``redo_lcia(demand)`` directly.

        """
        assert hasattr(self, "matrix"), "Must do `calculate` first"
        if demand is not None:
            self.build_demand_array(demand)
        self.calculate_scores()
        self.logger.info("Redoing LCI", extra={'demand': str(demand or self.demand)})



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
from eight import *

from .errors import AllArraysEmpty
from pathlib import Path
import datetime
import hashlib
import json
import numpy as np
import os
import tarfile
import tempfile

MAX_INT_32 = 4294967295
MAX_SIGNED_INT_32 = 2147483647


def load_arrays(objs):
    """Load the numpy arrays from list of objects ``objs``.

    Currently accepts ``str`` filepaths, ``BytesIO``,
     ``numpy.ndarray`` arrays. Creates copies of objects"""

    arrays = []

    is_array = lambda x: isinstance(x, np.ndarray)
    is_filepath = lambda x: isinstance(x, (str, Path))
    is_other = lambda x: not is_array(x) and not is_filepath(x)

    for obj in (o for o in objs if is_array(o)):
        arrays.append(obj.copy())

    for obj in sorted([o for o in objs if is_filepath(o)]):
        arrays.append(np.load(obj, allow_pickle=True))

    for obj in sorted([o for o in objs if is_other(o)]):
        arrays.append(np.load(obj, allow_pickle=True))

    if all(arr.shape[0] == 0 for arr in arrays):
        raise AllArraysEmpty

    return np.hstack(arrays)


def extract_uncertainty_fields(array):
    """Extract the core set of fields needed for uncertainty analysis from a parameter array"""
    fields = ["uncertainty_type", "amount", 'loc',
              'scale', 'shape', 'minimum', 'maximum', 'negative']
    return array[fields].copy()


try:
    from bw2data import (
        config,
        Database,
        databases,
        geomapping,
        mapping,
        Method,
        methods,
        Normalization,
        normalizations,
        projects,
        Weighting,
        weightings,
    )

    from bw2data.utils import TYPE_DICTIONARY, safe_filename

    TYPE_DICTIONARY["generic production"] = 11
    TYPE_DICTIONARY["generic consumption"] = 12

    global_index = geomapping[config.global_location]

    # Extension packages should extend OBJECT_MAPPING
    # with their additional classes

    OBJECT_MAPPING = {
        'database': (Database, databases),
        'method': (Method, methods),
        'normalization': (Normalization, normalizations),
        'weighting': (Weighting, weightings),
    }

    def get_database_filepath(functional_unit):
        """Get filepaths for all databases in supply chain of `functional_unit`"""
        dbs = set.union(*[Database(key[0]).find_graph_dependents() for key in functional_unit])
        return [Database(obj).filepath_processed() for obj in dbs]

    def get_filepaths(name, kind):
        """Get filepath for datastore object `name` of kind `kind`"""
        if kind == "demand":
            return get_database_filepath(name)
        if name is None:
            return None
        data_store, metadata = OBJECT_MAPPING[kind]
        assert name in metadata, "Can't find {} object {}".format(kind, name)
        return [data_store(name).filepath_processed()]

    def clean_databases():
        databases.clean()

    def save_calculation_package(name, demand, **kwargs):
        """Save a calculation package for later use in an independent LCA.

        Args:
            * name (str): Name of file to create. Will have datetime appended.
            * demand (dict): Demand dictionary.
            * kwargs: Any additional keyword arguments, e.g. ``method``, ``iterations``...

        Returns the filepath of the calculation package archive.

        """
        _ = lambda x: [os.path.basename(y) for y in x]

        filepaths = get_filepaths(demand, "demand")
        data = {
            'demand': {mapping[k]: v for k, v in demand.items()},
            'database_filepath': _(filepaths),
            'adjust_filepaths': ['database_filepath'],
        }
        for key, value in kwargs.items():
            if key in OBJECT_MAPPING:
                data[key] = _(get_filepaths(value, key))
                data['adjust_filepaths'].append(key)
                filepaths.extend(get_filepaths(value, key))
            else:
                data[key] = value

        config_fp = os.path.join(
            projects.output_dir,
            "{}.config.json".format(safe_filename(name))
        )
        with open(config_fp, "w") as f:
            json.dump(data, f, indent=2, ensure_ascii=False)
        archive = os.path.join(
            projects.output_dir,
            "{}.{}.tar.gz".format(
                safe_filename(name, False),
                datetime.datetime.now().strftime("%d-%B-%Y-%I-%M%p")
            )
        )
        with tarfile.open(archive, "w:gz") as tar:
            tar.add(
                config_fp,
                arcname=os.path.basename(config_fp)
            )
            for filepath in filepaths:
                tar.add(
                    filepath,
                    arcname=os.path.basename(filepath)
                )

        os.remove(config_fp)
        return archive
except ImportError:
    get_filepaths = mapping = global_index = None

    def clean_databases():
        pass

    def save_calculation_package(*arg, **kwargs):
        raise NotImplemented

    # Maximum value for unsigned integer stored in 4 bytes
    TYPE_DICTIONARY = {
        "unknown": -1,
        "production": 0,
        "technosphere": 1,
        "biosphere": 2,
        "substitution": 3,
        # Added for single-matrix LCAs, which combine LCI and LCIA.
        # consumption values need to be multiplied by -1 in the matrix.
        "generic production": 11,
        "generic consumption": 12,
    }


def get_seed(seed=None):
    """Get valid Numpy random seed value"""
    # https://groups.google.com/forum/#!topic/briansupport/9ErDidIBBFM
    random = np.random.RandomState(seed)
    return random.randint(0, MAX_SIGNED_INT_32)


def load_calculation_package(fp):
    """Load a calculation package created by ``save_calculation_package``.

    NumPy arrays are saved to a temporary directory, and file paths are adjusted.

    ``fp`` is the absolute file path of a calculation package file.

    Returns a dictionary suitable for passing to an LCA object, e.g. ``LCA(**load_calculation_package(fp))``.

    """
    assert os.path.exists(fp), "Can't find file: {}".format(fp)

    temp_dir = tempfile.mkdtemp()
    with tarfile.open(fp, 'r|gz') as tar:
        tar.extractall(temp_dir)

    config_fps = [x for x in os.listdir(temp_dir) if x.endswith(".config.json")]
    assert len(config_fps) == 1, "Can't find configuration file"
    config = json.load(open(os.path.join(temp_dir, config_fps[0])))
    config['demand'] = {int(k): v for k, v in config['demand'].items()}

    for field in config.pop('adjust_filepaths'):
        config[field] = [os.path.join(temp_dir, fn) for fn in config[field]]

    return config


def wrap_functional_unit(dct):
    """Transform functional units for effective logging.

    Turns ``Activity`` objects into their keys."""
    data = []
    for key, amount in dct.items():
        try:
            data.append({'database': key[0],
                         'code': key[1],
                         'amount': amount})
        except:
            data.append({'key': key,
                         'amount': amount})
    return data


def md5(filepath, blocksize=65536):
    """Generate MD5 hash for file at `filepath`"""
    hasher = hashlib.md5()
    fo = open(filepath, 'rb')
    buf = fo.read(blocksize)
    while len(buf) > 0:
        hasher.update(buf)
        buf = fo.read(blocksize)
    return hasher.hexdigest()



from datetime import datetime
import json
import logging
import os
import uuid

"""Adapted from json-log-formatter (https://github.com/marselester/json-log-formatter)

The MIT License (MIT)

Copyright (c) 2015 Marsel Mavletkulov

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""

BUILTIN_ATTRS = {
    'args',
    'asctime',
    'created',
    'exc_info',
    'exc_text',
    'filename',
    'funcName',
    'levelname',
    'levelno',
    'lineno',
    'module',
    'msecs',
    'message',
    'msg',
    'name',
    'pathname',
    'process',
    'processName',
    'relativeCreated',
    'stack_info',
    'thread',
    'threadName',
}


class JSONFormatter(logging.Formatter):
    json_lib = json

    def format(self, record):
        message = record.getMessage()
        extra = self.extra_from_record(record)
        json_record = self.json_record(message, extra, record)
        return self.to_json(json_record)

    def to_json(self, record):
        return self.json_lib.dumps(record, ensure_ascii=False)

    def extra_from_record(self, record):
        return {
            attr_name: record.__dict__[attr_name]
            for attr_name in record.__dict__
            if attr_name not in BUILTIN_ATTRS
        }

    def json_record(self, message, extra, record):
        if extra:
            data = extra
            if message:
                data['message'] = message
        else:
            # Also support logger.info({'foo': 'bar'})
            data = eval(message)
        data['time'] = str(datetime.utcnow())

        if record.exc_info:
            data['exc_info'] = self.formatException(record.exc_info)
        return data


def create_logger(dirpath=None, name=None, **kwargs):
    """Create a ``logger`` instance named ``bw2calc`` that can be used to log calculations.

    ``dirpath`` is the directory where the log file is saved. If ``dirpath`` is ``None``, no logger is created.

    ``name`` is the name of the calculation run, used to construct the log filepath.

    You can add other types of loggers, just add another handler to the ``bw2calc`` named logger before starting your calculations.

    Returns the filepath of the created log file.

    TODO: Decide on whether we copy safe_filepath to this package or create a common core package."""
    if dirpath is None:
        return

    assert os.path.isdir(dirpath) and os.access(dirpath, os.W_OK)

    # Use safe_filepath here
    filename = "{}.{}.json".format(
        name or uuid.uuid4().hex,
        str(datetime.utcnow())
    )

    formatter = JSONFormatter()
    fp = os.path.abspath(os.path.join(dirpath, filename))

    json_handler = logging.FileHandler(filename=fp, encoding='utf-8')
    json_handler.setFormatter(formatter)

    logger = logging.getLogger('bw2calc')
    logger.addHandler(json_handler)
    logger.setLevel(logging.INFO)

    return fp



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
from eight import *

from bw2calc import *
from bw2data import config, Database, Method, projects, databases
from bw2data.utils import random_string
from bw2data.tests import bw2test
from numbers import Number
import numpy as np
import os
import pytest
import wrapt


no_pool = pytest.mark.skipif(config._windows,
    reason="fork() on Windows doesn't pass temp directory")

yes_docker = pytest.mark.skipif(bool(os.environ.get("BRIGHTWAY2_DOCKER")),
    reason="Project directory in CI Docker container is '/home/jovyan/data'")
no_docker = pytest.mark.skipif(not bool(os.environ.get("BRIGHTWAY2_DOCKER")),
    reason="Normal project directory")


def build_databases():
    Database("biosphere").write({
        ("biosphere", "1"): {'type': 'emission'},
        ("biosphere", "2"): {'type': 'emission'},
    })
    Database("test").write({
        ("test", "1"): {
            'exchanges': [{
                'amount': 0.5,
                'minimum': 0.2,
                'maximum': 0.8,
                'input': ('test', "2"),
                'type': 'technosphere',
                'uncertainty type': 4
            }, {
                'amount': 1,
                'minimum': 0.5,
                'maximum': 1.5,
                'input': ('biosphere', "1"),
                'type': 'biosphere',
                'uncertainty type': 4
            }],
            'type': 'process',
        },
        ("test", "2"): {
            'exchanges': [{
                'amount': 0.1,
                'minimum': 0,
                'maximum': 0.2,
                'input': ('biosphere', "2"),
                'type': 'biosphere',
                'uncertainty type': 4
            }],
            'type': 'process',
            'unit': 'kg'
        },
    })
    method = Method(("a", "method"))
    method.register()
    method.write([
        (("biosphere", "1"), 1),
        (("biosphere", "2"), 2),
    ])


@pytest.fixture
@bw2test
def background():
    build_databases()


@wrapt.decorator
def random_project(wrapped, instance, args, kwargs):
    config.is_test = True
    projects._restore_orig_directory()
    string = random_string()
    while string in projects:
        string = random_string()
    projects.set_current(string)
    build_databases()
    result = wrapped(*args, **kwargs)
    projects.set_current("default", writable=False)
    projects.delete_project(name=string, delete_dir=True)
    return result


def get_args():
    return {("test", "1"): 1}, ("a", "method")


@random_project
@yes_docker
def test_random_project():
    print(projects.dir)
    assert "Brightway" in projects.dir

@random_project
@no_docker
def test_random_project():
    print(projects.dir)
    assert "jovyan" in projects.dir

@bw2test
def test_temp_dir_again():
    assert "Brightway" not in projects.dir

def test_plain_monte_carlo(background):
    mc = MonteCarloLCA(*get_args())
    if hasattr(mc, "__next__"):
        assert mc.__next__() > 0
    else:
        assert mc.next() > 0

def test_monte_carlo_next(background):
    mc = MonteCarloLCA(*get_args())
    assert next(mc) > 0

def test_monte_carlo_as_iterator(background):
    mc = MonteCarloLCA(*get_args())
    for x in mc:
        assert x > 0
        break

def test_direct_solving(background):
    mc = DirectSolvingMonteCarloLCA(*get_args())
    assert next(mc)

def test_parameter_vector_monte_carlo(background):
    mc = ParameterVectorLCA(*get_args())
    assert next(mc) > 0

@no_pool
def test_multi_mc(background):
    mc = MultiMonteCarlo(
        [
            {("test", "1"): 1},
            {("test", "2"): 1},
            {("test", "1"): 1, ("test", "2"): 1}
        ],
        ("a", "method"),
        iterations=10
    )
    results = mc.calculate()
    assert results

@no_pool
def test_multi_mc_not_same_answer(background):
    activity_list = [
            {("test", "1"): 1},
            {("test", "2"): 1},
            # {("test", "1"): 1, ("test", "2"): 1}
        ]
    mc = MultiMonteCarlo(
        activity_list,
        ("a", "method"),
        iterations=10
    )
    results = mc.calculate()
    assert len(results) == 2
    for _, lst in results:
        assert len(set(lst)) == len(lst)

    lca = LCA(activity_list[0], ("a", "method"))
    lca.lci()
    lca.lcia()

    def score(lca, func_unit):
        lca.redo_lcia(func_unit)
        return lca.score

    static = [score(lca, func_unit) for func_unit in activity_list]
    for a, b in zip(static, results):
        assert a not in b[1]

@no_pool
def test_multi_mc_compound_func_units(background):
    activity_list = [
            {("test", "1"): 1},
            {("test", "2"): 1},
            {("test", "1"): 1, ("test", "2"): 1}
        ]
    mc = MultiMonteCarlo(
        activity_list,
        ("a", "method"),
        iterations=10
    )
    results = mc.calculate()
    assert len(results) == 3
    assert activity_list == mc.demands

@random_project
def test_multi_mc_no_temp_dir():
    mc = MultiMonteCarlo(
        [
            {("test", "1"): 1},
            {("test", "2"): 1},
            {("test", "1"): 1, ("test", "2"): 1}
        ],
        ("a", "method"),
        iterations=10
    )
    results = mc.calculate()
    print(results)
    assert results
    assert isinstance(results, list)
    assert len(results)

@no_pool
def test_parallel_monte_carlo(background):
    fu, method = get_args()
    mc = ParallelMonteCarlo(fu, method, iterations=200)
    results = mc.calculate()
    print(results)
    assert results

@random_project
def test_parallel_monte_carlo_no_temp_dir():
    fu, method = get_args()
    mc = ParallelMonteCarlo(fu, method, iterations=200)
    results = mc.calculate()
    print(results)
    assert results
    assert isinstance(results, list)
    assert isinstance(results[0], Number)
    assert results[0] > 0



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
from eight import *

from bw2calc import *
from bw2data import Database, Method, mapping
from bw2data.tests import bw2test
import numpy as np
import pytest


def build_databases():
    Database("biosphere").write({
        ("biosphere", "1"): {'type': 'emission'},
        ("biosphere", "2"): {'type': 'emission'},
    })
    Database("test").write({
        ("test", "1"): {
            'exchanges': [{
                'amount': 0.5,
                'minimum': 0.2,
                'maximum': 0.8,
                'input': ('test', "2"),
                'type': 'technosphere',
                'uncertainty type': 4  # Uniform
            }, {
                'amount': 100,
                'minimum': 50,
                'maximum': 500,
                'input': ('biosphere', "1"),
                'type': 'biosphere',
                'loc': 100,
                'scale': 20,
                'uncertainty type': 3  # Normal
            }],
            'type': 'process',
        },
        ("test", "2"): {
            'exchanges': [{
                'amount': -0.42,
                'input': ('biosphere', "2"),
                'type': 'biosphere',
            }],
            'type': 'process',
            'unit': 'kg'
        },
    })
    method = Method(("a", "method"))
    method.register()
    method.write([
        (("biosphere", "1"), 1),
        (("biosphere", "2"), {
            'amount': 10,
            'uncertainty type': 5,  # Triangular
            'loc': 10,
            'minimum': 8,
            'maximum': 15
        }),
    ])


@pytest.fixture
@bw2test
def background():
    build_databases()


def test_pv_no_need_load_data(background):
    assert next(ParameterVectorLCA({("test", "1"): 1}, ("a", "method")))


def test_pv_ordering_correct(background):
    pv = ParameterVectorLCA({("test", "1"): 1}, ("a", "method"))
    pv.load_data()

    b1 = pv.biosphere_dict[("biosphere", "1")]
    b2 = pv.biosphere_dict[("biosphere", "2")]
    p1 = pv.product_dict[("test", "1")]
    p2 = pv.product_dict[("test", "2")]
    a1 = pv.activity_dict[("test", "1")]
    a2 = pv.activity_dict[("test", "2")]

    for _ in range(10):
        next(pv)
        assert pv.characterization_matrix[b1, b1] == 1
        assert 8 <= pv.characterization_matrix[b2, b2] <= 15
        assert pv.characterization_matrix[b2, b1] == 0

        assert np.allclose(pv.biosphere_matrix[b2, a2], -0.42)
        assert pv.biosphere_matrix[b1, a1] >= 50

        assert pv.technosphere_matrix[p1, a1] == 1
        assert pv.technosphere_matrix[p2, a2] == 1
        assert pv.technosphere_matrix[p1, a2] == 0
        # Inputs are negative
        assert -0.8 <= pv.technosphere_matrix[p2, a1] <= -0.2

        assert 8 <= pv.characterization_matrix[b2, b2] <= 15
        assert pv.characterization_matrix[b2, b1] == 0


def test_raise_assertion_error_with_wrong_size_vector(background):
    pv = ParameterVectorLCA({("test", "1"): 1}, ("a", "method"))
    with pytest.raises(AssertionError):
        pv.rebuild_all(np.ones(100,))


def test_no_error_with_right_size_vector(background):
    pv = ParameterVectorLCA({("test", "1"): 1}, ("a", "method"))
    pv.rebuild_all(np.ones(7,))


def test_stored_samples_correct(background):
    pv = ParameterVectorLCA({("test", "1"): 1}, ("a", "method"))
    pv.load_data()

    samples = []
    for x in range(10):
        next(pv)
        samples.append(pv.sample.copy())

    expected = [
        (mapping[("test", "2")], mapping[("test", "2")], 1, 1),
        (mapping[('test', "1")], mapping[("test", "1")], 1, 1),
        (mapping[('test', "2")], mapping[("test", "1")], 0.2, 0.8),
        (mapping[('biosphere', "1")], mapping[("test", "1")], 50, 500),
        (mapping[('biosphere', "2")], mapping[("test", "2")], -0.43, -0.41),
    ]

    concatenated = np.hstack((pv.tech_params, pv.bio_params))
    for row, col, mn, mx in expected:
        mask = (concatenated['input'] == row) * (concatenated['output'] == col)
        assert mask.sum() == 1
        for sample in samples:
            assert mn <= sample[:mask.shape[0]][mask] <= mx



import numpy as N
from scipy import sparse
# Modified from http://www.scipy.org/scipy/scipy/attachment/ticket/602/merge_sparse_blocks.py

def merge_sparse_blocks(block_mats, lookup_dicts, format='csr', dtype=N.float64):
    """.. centered:: Merge several sparse matrix blocks into a single sparse matrix

Input Params
============

block_mats -- Iterable of sparse matrices
lookup_dicts -- Iterable of lookup dictionaries (arbitrary value -> array index)
format -- Desired sparse format of output matrix
dtype -- Desired dtype, defaults to N.float64

Output
======

Merged matrix, to array indices dictionary, from array indices dictionary

Example
=======

Combine two matrices, and check that they are combined correctly, and that the lookup dictionaries are correct as well.

Matrix a:   [[0  1   2   3]
            [1  1   0   0]
            [2  0   2   0]
            [3  0   0   3]]

Matrix b:   [[4   5   6   7]
            [5  5   0   0]
            [6  0   6   0]
            [7  0   0   7]]

Lookup dict a: {20: 0, 21: 1, 22: 2, 23: 3}

Lookup dict b: {30: 0, 31: 1, 32: 2, 33: 3}

    >>> a = sparse.lil_matrix((4,4))
    >>> a[:,0] = range(4)
    >>> a[0,:] = range(4)
    >>> a.setdiag(range(4))

    >>> b = sparse.lil_matrix((4,4))
    >>> b[:,0] = range(4,8)
    >>> b[0,:] = range(4,8)
    >>> b.setdiag(range(4,8))

    >>> lookup_a = {20: 0, 21: 1, 22: 2, 23: 3}
    >>> lookup_b = {30: 0, 31: 1, 32: 2, 33: 3}

    >>> c, forwards, backwards = merge_sparse_blocks((a,b), (lookup_a, lookup_b))
    >>> assert isinstance(c, sparse.sparse.csr_matrix)
    >>> print forwards[20]
    0
    >>> print forwards[30]
    4
    >>> print backwards[1]
    21
    >>> print backwards[5]
    31
    >>> print N.array(c.todense())
    [[ 0.  1.  2.  3.  0.  0.  0.  0.]
     [ 1.  1.  0.  0.  0.  0.  0.  0.]
     [ 2.  0.  2.  0.  0.  0.  0.  0.]
     [ 3.  0.  0.  3.  0.  0.  0.  0.]
     [ 0.  0.  0.  0.  4.  5.  6.  7.]
     [ 0.  0.  0.  0.  5.  5.  0.  0.]
     [ 0.  0.  0.  0.  6.  0.  6.  0.]
     [ 0.  0.  0.  0.  7.  0.  0.  7.]]
    """
    def cumulative_sum(l):
         cum = list(l[:]) # Change if tuple to allow assignment
         csum_0 = 0
         csum_1 = 0
         csum_2 = 0
         for i, (x, y, z) in enumerate(l):
             csum_0 += x
             csum_1 += y
             csum_2 += z
             cum[i] = (csum_0, csum_1, csum_2)
         return cum

    block_mats = [mat.tocoo() for mat in block_mats]
    # Set up a triple of x offset, y offset, nnz (data length) offset
    offsets = cumulative_sum([(0,0,0)] + [(mat.shape[0], mat.shape[1], mat.nnz) for mat in block_mats]) # Last pair is not an offset, but final shape
    final_shape = offsets[-1]
    data = N.empty(final_shape[2], dtype=dtype) # Values in matrix
    row = N.empty(final_shape[2], dtype=N.intc) # Row indices
    col = N.empty(final_shape[2], dtype=N.intc) # Col indices
    for index, bm in enumerate(block_mats):
        data[offsets[index][2]:offsets[index+1][2]] = bm.data
        row[offsets[index][2]:offsets[index+1][2]] = bm.row+offsets[index][0]
        col[offsets[index][2]:offsets[index+1][2]] = bm.col+offsets[index][1] # Check x and y even though matrices are square

    merged_mat = sparse.coo_matrix((data, (row, col)), dtype=dtype)

    if format != 'coo':
        merged_mat = getattr(merged_mat, 'to'+format)()
        # if format == 'csc' or format == 'csr':
        #     if not merged_mat.has_sorted_indices: merged_mat.sort_indices()

    values = N.empty(final_shape[0], dtype=N.intc)
    keys = N.empty(final_shape[0], dtype=N.intc)
    for index, dic in enumerate(lookup_dicts):
        values[offsets[index][0]:offsets[index+1][0]] = N.array(dic.values()) + offsets[index][0]
        keys[offsets[index][0]:offsets[index+1][0]] = dic.keys()

    lookup_forward_dict = dict(zip(keys, values))
    lookup_backwards_dict = dict(zip(values, keys))

    return merged_mat, lookup_forward_dict, lookup_backwards_dict # Forward is into array

def merge_sparse_proxies(first, second):
    pass

def _test():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    _test()